"""Test cases for pycatools."""

import datetime
import math
import threading
import time
import unittest

from pycatools import AlarmStatus
from pycatools import AlarmSeverity
from pycatools import CaValue
from pycatools import PvValue
from pycatools import PvValueMatcher
from pycatools import CagetParser
from pycatools import CagetStderrAnalyzer
from pycatools import CaputParser
from pycatools import CaputStderrAnalyzer
from pycatools import CamonitorParser
from pycatools import Caget
from pycatools import Caput
from pycatools import Camonitor
from pycatools import Pv
from pycatools import TimeoutError

def suite():
  result = unittest.TestSuite()
  result.addTest(unittest.makeSuite(AlarmStatusTest))
  result.addTest(unittest.makeSuite(AlarmSeverityTest))
  result.addTest(unittest.makeSuite(CaValueTest))
  result.addTest(unittest.makeSuite(PvValueMatcherTest))
  result.addTest(unittest.makeSuite(CagetParserTest))
  result.addTest(unittest.makeSuite(CagetStderrAnalyzerTest))
  result.addTest(unittest.makeSuite(CaputParserTest))
  result.addTest(unittest.makeSuite(CaputStderrAnalyzerTest))
  result.addTest(unittest.makeSuite(CamonitorParserTest))
  result.addTest(unittest.makeSuite(CagetTest))
  result.addTest(unittest.makeSuite(CaputTest))
  result.addTest(unittest.makeSuite(CamonitorTest))
  result.addTest(unittest.makeSuite(PvTest))
  return result

class CaputSequenceThread(threading.Thread):
  """I am a Thread who performs a sequence of CA puts."""

  # @protocol("initialize-release")

  def __init__(self, aListOfNameValueSleepLists, timeoutNumberInSeconds=None):
    """A None name or value indicates a no-op."""
    threading.Thread.__init__(self)
    self.daemon = True
    self.nameValueSleeps = aListOfNameValueSleepLists
    self.timeout = timeoutNumberInSeconds
    self.stopRequested = threading.Event()

  # @protocol("running")

  def run(self):
    for each in self.nameValueSleeps:
      if self.stopRequested.is_set(): break
      (pvName, value, sleepTime) = each
      if sleepTime != None and sleepTime != 0.0: time.sleep(sleepTime)
      if pvName != None and value != None:
        caput = Caput(pvName, value)
        caput.setBlocking(True)
        if self.timeout != None: caput.setTimeout(self.timeout)
        caput.run()

  # @protocol("stopping")

  def requestStop(self):
    self.stopRequested.set()

class AlarmStatusTest(unittest.TestCase):
  def testAsString01(self):
    self.assertEqual("NO_ALARM", AlarmStatus.NO_ALARM.asString())
    self.assertEqual("READ", AlarmStatus.READ.asString())

class AlarmSeverityTest(unittest.TestCase):
  def testAsString01(self):
    self.assertEqual("NO_ALARM", AlarmSeverity.NO_ALARM.asString())
    self.assertEqual("MINOR", AlarmSeverity.MINOR.asString())

class CaValueTest(unittest.TestCase):
  def testIsValid01(self):
    self.assertFalse(CaValue().isValid())
    self.assertFalse(CaValue(severity=None).isValid())
    self.assertFalse(CaValue(severity=AlarmSeverity.INVALID).isValid())
    self.assertTrue(CaValue(severity=AlarmSeverity.MAJOR).isValid())

class PvValueMatcherTest(unittest.TestCase):
  def testMatches01(self):
    """Matches all."""
    now = datetime.datetime.now()
    pvValue = PvValue("ioc23:val1", 3, now, AlarmStatus.NO_ALARM,
        AlarmSeverity.NO_ALARM)
    matcher = PvValueMatcher("ioc23:val1", 3, None, now, AlarmStatus.NO_ALARM,
        AlarmSeverity.NO_ALARM)
    self.assertTrue(matcher.matches(pvValue))

  def testMatches02(self):
    """PV name does not match."""
    now = datetime.datetime.now()
    pvValue = PvValue("ioc23:val1", 3, now, AlarmStatus.NO_ALARM,
        AlarmSeverity.NO_ALARM)
    matcher = PvValueMatcher("ioc23:val2", 3, None, now, AlarmStatus.NO_ALARM,
        AlarmSeverity.NO_ALARM)
    self.assertFalse(matcher.matches(pvValue))

  def testMatches03(self):
    """Value does not match."""
    now = datetime.datetime.now()
    pvValue = PvValue("ioc23:val1", 3, now, AlarmStatus.NO_ALARM,
        AlarmSeverity.NO_ALARM)
    matcher = PvValueMatcher("ioc23:val1", 4, None, now, AlarmStatus.NO_ALARM,
        AlarmSeverity.NO_ALARM)
    self.assertFalse(matcher.matches(pvValue))

  def testMatches04(self):
    """Timestamp does not match."""
    ts1 = datetime.datetime(2010, 7, 22, 15, 55, 9, 615251)
    ts2 = datetime.datetime(2010, 7, 22, 15, 56, 9, 615251)
    pvValue = PvValue("ioc23:val1", 3, ts1, AlarmStatus.NO_ALARM,
        AlarmSeverity.NO_ALARM)
    matcher = PvValueMatcher("ioc23:val1", 3, None, ts2, AlarmStatus.NO_ALARM,
        AlarmSeverity.NO_ALARM)
    self.assertFalse(matcher.matches(pvValue))

  def testMatches05(self):
    """Status does not match."""
    now = datetime.datetime.now()
    pvValue = PvValue("ioc23:val1", 3, now, AlarmStatus.NO_ALARM,
        AlarmSeverity.NO_ALARM)
    matcher = PvValueMatcher("ioc23:val1", 3, None, now, AlarmStatus.READ,
        AlarmSeverity.NO_ALARM)
    self.assertFalse(matcher.matches(pvValue))

  def testMatches06(self):
    """Severity does not match."""
    now = datetime.datetime.now()
    pvValue = PvValue("ioc23:val1", 3, now, AlarmStatus.NO_ALARM,
        AlarmSeverity.NO_ALARM)
    matcher = PvValueMatcher("ioc23:val1", 3, None, now, AlarmStatus.NO_ALARM,
        AlarmSeverity.MINOR)
    self.assertFalse(matcher.matches(pvValue))

  def testMatches07(self):
    """PV name don't care."""
    now = datetime.datetime.now()
    pvValue = PvValue("ioc23:val1", 3, now, AlarmStatus.NO_ALARM,
        AlarmSeverity.NO_ALARM)
    matcher = PvValueMatcher(None, 3, None, now, AlarmStatus.NO_ALARM,
        AlarmSeverity.NO_ALARM)
    self.assertTrue(matcher.matches(pvValue))

  def testMatches08(self):
    """Value don't care."""
    now = datetime.datetime.now()
    pvValue = PvValue("ioc23:val1", 3, now, AlarmStatus.NO_ALARM,
        AlarmSeverity.NO_ALARM)
    matcher = PvValueMatcher("ioc23:val1", None, None, now,
        AlarmStatus.NO_ALARM, AlarmSeverity.NO_ALARM)
    self.assertTrue(matcher.matches(pvValue))

  def testMatches09(self):
    """Timestamp don't care."""
    ts1 = datetime.datetime(2010, 7, 22, 15, 55, 9, 615251)
    pvValue = PvValue("ioc23:val1", 3, ts1, AlarmStatus.NO_ALARM,
        AlarmSeverity.NO_ALARM)
    matcher = PvValueMatcher("ioc23:val1", 3, None, None,
        AlarmStatus.NO_ALARM, AlarmSeverity.NO_ALARM)
    self.assertTrue(matcher.matches(pvValue))

  def testMatches10(self):
    """Status don't care."""
    now = datetime.datetime.now()
    pvValue = PvValue("ioc23:val1", 3, now, AlarmStatus.NO_ALARM,
        AlarmSeverity.NO_ALARM)
    matcher = PvValueMatcher("ioc23:val1", 3, None, now, None,
        AlarmSeverity.NO_ALARM)
    self.assertTrue(matcher.matches(pvValue))

  def testMatches11(self):
    """Severity don't care."""
    now = datetime.datetime.now()
    pvValue = PvValue("ioc23:val1", 3, now, AlarmStatus.NO_ALARM,
        AlarmSeverity.NO_ALARM)
    matcher = PvValueMatcher("ioc23:val1", 3, None, now, AlarmStatus.NO_ALARM,
        None)
    self.assertTrue(matcher.matches(pvValue))

  def testMatches12(self):
    """PV name function."""
    now = datetime.datetime.now()
    pvValue = PvValue("ioc23:val1", 3, now, AlarmStatus.NO_ALARM,
        AlarmSeverity.NO_ALARM)
    matcher = PvValueMatcher(
        lambda aPvValue: aPvValue.getPvName().upper() == "IOC23:VAL1", 3,
        None, now, AlarmStatus.NO_ALARM, AlarmSeverity.NO_ALARM)
    self.assertTrue(matcher.matches(pvValue))

  def testMatches13(self):
    """Value function."""
    now = datetime.datetime.now()
    pvValue = PvValue("ioc23:val1", 3, now, AlarmStatus.NO_ALARM,
        AlarmSeverity.NO_ALARM)
    matcher = PvValueMatcher("ioc23:val1",
        lambda aPvValue: aPvValue.getValue() > 2, None, now,
        AlarmStatus.NO_ALARM, AlarmSeverity.NO_ALARM)
    self.assertTrue(matcher.matches(pvValue))

  def testMatches14(self):
    """Timestamp function."""
    ts1 = datetime.datetime(2010, 7, 22, 15, 55, 9, 615251)
    ts2 = datetime.datetime(2010, 7, 22, 10, 55, 9, 615251)
    pvValue = PvValue("ioc23:val1", 3, ts1, AlarmStatus.NO_ALARM,
        AlarmSeverity.NO_ALARM)
    matcher = PvValueMatcher("ioc23:val1", 3, None,
        lambda aPvValue: aPvValue.getTimestamp() > ts2, AlarmStatus.NO_ALARM,
        AlarmSeverity.NO_ALARM)
    self.assertTrue(matcher.matches(pvValue))

  def testMatches14(self):
    """Status function."""
    now = datetime.datetime.now()
    pvValue = PvValue("ioc23:val1", 3, now, AlarmStatus.NO_ALARM,
        AlarmSeverity.NO_ALARM)
    matcher = PvValueMatcher("ioc23:val1", 3, None, now,
        lambda aPvValue: aPvValue.getStatus() == AlarmStatus.NO_ALARM,
        AlarmSeverity.NO_ALARM)
    self.assertTrue(matcher.matches(pvValue))

  def testMatches15(self):
    """Severity function."""
    now = datetime.datetime.now()
    pvValue = PvValue("ioc23:val1", 3, now, AlarmStatus.NO_ALARM,
        AlarmSeverity.NO_ALARM)
    matcher = PvValueMatcher("ioc23:val1", 3, None, now, AlarmStatus.NO_ALARM,
        lambda aPvValue: aPvValue.getSeverity() == AlarmSeverity.NO_ALARM)
    self.assertTrue(matcher.matches(pvValue))

  def testMatches16(self):
    """Float value within delta."""
    now = datetime.datetime.now()
    pvValue = PvValue("ioc23:val1", 3.123, now, AlarmStatus.NO_ALARM,
        AlarmSeverity.NO_ALARM)
    matcher = PvValueMatcher("ioc23:val1", 3.122, 0.0011, now,
        AlarmStatus.NO_ALARM, AlarmSeverity.NO_ALARM)
    self.assertTrue(matcher.matches(pvValue))

  def testMatches17(self):
    """Long value."""
    now = datetime.datetime.now()
    pvValue = PvValue("ioc23:val1", 5, now, AlarmStatus.NO_ALARM,
        AlarmSeverity.NO_ALARM)
    matcher = PvValueMatcher("ioc23:val1", 5, None, now,
        AlarmStatus.NO_ALARM, AlarmSeverity.NO_ALARM)
    self.assertTrue(matcher.matches(pvValue))

class CagetParserTest(unittest.TestCase):
  def createParser(self, aString, fieldSeparatorString=","):
    return CagetParser(aString, fieldSeparatorString)

  def parse(self, aString, fieldSeparatorString=","):
    return self.createParser(aString, fieldSeparatorString).parse()

  def testParse01(self):
    """Normal."""
    result = self.parse(
        "x7ib:DS:mhl:HmCplt,2010-07-22 15:55:09.615251,1,,")
    self.assertEqual(1, len(result))
    self.assertEqual("x7ib:DS:mhl:HmCplt", result[0].getPvName())
    self.assertEqual(1, result[0].getValue())
    self.assertEqual(datetime.datetime(2010, 7, 22, 15, 55, 9, 615251),
        result[0].getTimestamp())
    self.assertEqual(AlarmStatus.NO_ALARM, result[0].getStatus())
    self.assertEqual(AlarmSeverity.NO_ALARM, result[0].getSeverity())

  def testParse02(self):
    """Non-NO_ALARM alarm status and severity."""
    result = self.parse(
        "x7ib:x:AmpEna,2010-07-22 17:23:51.395512,0,STATE,MAJOR")
    self.assertEqual(1, len(result))
    self.assertEqual("x7ib:x:AmpEna", result[0].getPvName())
    self.assertEqual(0, result[0].getValue())
    self.assertEqual(datetime.datetime(2010, 7, 22, 17, 23, 51, 395512),
        result[0].getTimestamp())
    self.assertEqual(AlarmStatus.STATE, result[0].getStatus())
    self.assertEqual(AlarmSeverity.MAJOR, result[0].getSeverity())

  def testParse03(self):
    """One PV not found and one found."""
    result = self.parse(
        "nonexistent,*** Not connected (PV not found)\n"
        "x7ib:DS:mhl:HmCplt,2010-07-22 15:55:09.615251,1,,")
    self.assertEqual(2, len(result))
    self.assertEqual("nonexistent", result[0].getPvName())
    self.assertEqual("x7ib:DS:mhl:HmCplt", result[1].getPvName())
    self.assertEqual(1, result[1].getValue())
    self.assertEqual(datetime.datetime(2010, 7, 22, 15, 55, 9, 615251),
        result[1].getTimestamp())
    self.assertEqual(AlarmStatus.NO_ALARM, result[1].getStatus())
    self.assertEqual(AlarmSeverity.NO_ALARM, result[1].getSeverity())

  def testParse04(self):
    """None field separator is converted to the default field separator."""
    result = self.parse(
        "x7ib:DS:mhl:HmCplt\0372010-07-22 15:55:09.615251\0371\037\037", None)
    self.assertEqual(1, len(result))
    self.assertEqual("x7ib:DS:mhl:HmCplt", result[0].getPvName())
    self.assertEqual(1, result[0].getValue())
    self.assertEqual(datetime.datetime(2010, 7, 22, 15, 55, 9, 615251),
        result[0].getTimestamp())
    self.assertEqual(AlarmStatus.NO_ALARM, result[0].getStatus())
    self.assertEqual(AlarmSeverity.NO_ALARM, result[0].getSeverity())

  def testParse05(self):
    """Two PVs."""
    result = self.parse(
        "ioc23:bo1,2010-07-22 15:55:09.615251,1,,\n"
        "ioc23:bo2,2010-07-23 15:55:09.615252,0,STATE,MAJOR")
    self.assertEqual(2, len(result))
    self.assertEqual("ioc23:bo1", result[0].getPvName())
    self.assertEqual(1, result[0].getValue())
    self.assertEqual(datetime.datetime(2010, 7, 22, 15, 55, 9, 615251),
        result[0].getTimestamp())
    self.assertEqual(AlarmStatus.NO_ALARM, result[0].getStatus())
    self.assertEqual(AlarmSeverity.NO_ALARM, result[0].getSeverity())
    self.assertEqual("ioc23:bo2", result[1].getPvName())
    self.assertEqual(0, result[1].getValue())
    self.assertEqual(datetime.datetime(2010, 7, 23, 15, 55, 9, 615252),
        result[1].getTimestamp())
    self.assertEqual(AlarmStatus.STATE, result[1].getStatus())
    self.assertEqual(AlarmSeverity.MAJOR, result[1].getSeverity())

  def testParse06(self):
    """No PVs."""
    result = self.parse("")
    self.assertEqual(0, len(result))

  def testParse07(self):
    """String value."""
    result = self.parse(
        "ioc23:bo1,2010-07-22 15:55:09.615251,abc,,")
    self.assertEqual("abc", result[0].getValue())

  def testParse08(self):
    """String value with a space."""
    result = self.parse(
        "ioc23:bo1,2010-07-22 15:55:09.615251,ab c,,")
    self.assertEqual("ab c", result[0].getValue())

  def testParse09(self):
    """Float value."""
    result = self.parse(
        "ioc23:bo1,2010-07-22 15:55:09.615251,1.23,,")
    self.assertAlmostEqual(1.23, result[0].getValue())

  def testParse10(self):
    """Float value in E notation."""
    result = self.parse(
        "ioc23:bo1,2010-07-22 15:55:09.615251,1.23e2,,")
    self.assertAlmostEqual(1.23e2, result[0].getValue())

  def testParse11(self):
    """Negative float value in E notation."""
    result = self.parse(
        "ioc23:bo1,2010-07-22 15:55:09.615251,-1.23e2,,")
    self.assertAlmostEqual(-1.23e2, result[0].getValue())

  def testParse12(self):
    """Negative float value in E notation with negative exponent."""
    result = self.parse(
        "ioc23:bo1,2010-07-22 15:55:09.615251,-1.23e-2,,")
    self.assertAlmostEqual(-1.23e-2, result[0].getValue())

  def testParse13(self):
    """Negative int value."""
    result = self.parse(
        "ioc23:bo1,2010-07-22 15:55:09.615251,-4,,")
    self.assertEqual(-4, result[0].getValue())

  def testParse14(self):
    """String starting with digits."""
    result = self.parse(
        "ioc23:bo1,2010-07-22 15:55:09.615251,123abc,,")
    self.assertEqual("123abc", result[0].getValue())

  def testParse15(self):
    """NaN."""
    result = self.parse(
        "ioc23:bo1,2010-07-22 15:55:09.615251,nan,,")
    self.assertTrue(math.isnan(result[0].getValue()))

  def testParse16(self):
    """Infinity."""
    result = self.parse(
        "ioc23:bo1,2010-07-22 15:55:09.615251,inf,,")
    self.assertEqual(float("inf"), result[0].getValue())

  def testParse17(self):
    """Undefined timestamp."""
    result = self.parse(
        "ioc23:val1,<undefined>,0,UDF,INVALID")
    self.assertEqual("ioc23:val1", result[0].getPvName())
    self.assertEqual(0, result[0].getValue())
    self.assertEqual(None, result[0].getTimestamp())
    self.assertEqual(AlarmStatus.UDF, result[0].getStatus())
    self.assertEqual(AlarmSeverity.INVALID, result[0].getSeverity())

  def testInstanceCreation01(self):
    """Multi-character field separator is not allowed."""
    self.assertRaises(ValueError, CagetParser, "", "::")

  def testInstanceCreation02(self):
    """Space field separator is not allowed."""
    self.assertRaises(ValueError, CagetParser, "", " ")

class CagetStderrAnalyzerTest(unittest.TestCase):
  def createAnalyzer(self, aString):
    return CagetStderrAnalyzer(aString)

  def testAnalyzer01(self):
    """Empty."""
    analyzer = self.createAnalyzer("")
    analyzer.analyze()
    self.assertFalse(analyzer.foundChannelConnectTimeout())

  def testAnalyzer02(self):
    """Channel connect timeout error."""
    analyzer = self.createAnalyzer(
        "Channel connect timed out: 'ioc23:nonexistent' not found.\n")
    analyzer.analyze()
    self.assertTrue(analyzer.foundChannelConnectTimeout())

class CaputParserTest(unittest.TestCase):
  def createParser(self, aString, fieldSeparatorString=","):
    return CaputParser(aString, fieldSeparatorString)

  def parse(self, aString, fieldSeparatorString=","):
    return self.createParser(aString, fieldSeparatorString).parse()

  def testParse01(self):
    """Normal."""
    result = self.parse(
        "Old : ioc23:val1,2012-09-27 18:23:13.013062,0,,\n"
        "New : ioc23:val1,2012-09-27 18:25:39.418092,1,,")
    self.assertEqual("ioc23:val1", result.getPvName())
    self.assertEqual(1, result.getValue())
    self.assertEqual(datetime.datetime(2012, 9, 27, 18, 25, 39, 418092),
        result.getTimestamp())
    self.assertEqual(AlarmStatus.NO_ALARM, result.getStatus())
    self.assertEqual(AlarmSeverity.NO_ALARM, result.getSeverity())

  def testParse02(self):
    """Non-NO_ALARM alarm status and severity."""
    result = self.parse(
        "Old : ioc23:val1,2012-09-27 18:23:13.013062,0,,\n"
        "New : ioc23:val1,2012-09-27 18:25:39.418092,1,STATE,MINOR")
    self.assertEqual("ioc23:val1", result.getPvName())
    self.assertEqual(1, result.getValue())
    self.assertEqual(datetime.datetime(2012, 9, 27, 18, 25, 39, 418092),
        result.getTimestamp())
    self.assertEqual(AlarmStatus.STATE, result.getStatus())
    self.assertEqual(AlarmSeverity.MINOR, result.getSeverity())

  def testParse03(self):
    """Empty program output."""
    result = self.parse("")
    self.assertEqual(None, result)

class CaputStderrAnalyzerTest(unittest.TestCase):
  def createAnalyzer(self, aString):
    return CaputStderrAnalyzer(aString)

  def testAnalyzer01(self):
    """Empty."""
    analyzer = self.createAnalyzer("")
    analyzer.analyze()
    self.assertFalse(analyzer.foundChannelConnectTimeout())
    self.assertFalse(analyzer.foundWriteCallbackTimeout())

  def testAnalyzer02(self):
    """Not a write callback timeout error."""
    analyzer = self.createAnalyzer(
        "Channel connect timed out: 'ioc23:motorRqsPos1.CALC' not found.\n")
    analyzer.analyze()
    self.assertFalse(analyzer.foundWriteCallbackTimeout())

  def testAnalyzer03(self):
    """Write callback timeout error."""
    analyzer = self.createAnalyzer(
        "Write callback operation timed out\n")
    analyzer.analyze()
    self.assertTrue(analyzer.foundWriteCallbackTimeout())

  def testAnalyzer04(self):
    """Channel connect timeout error."""
    analyzer = self.createAnalyzer(
        "Channel connect timed out: 'ioc23:nonexistent' not found.\n")
    analyzer.analyze()
    self.assertTrue(analyzer.foundChannelConnectTimeout())

class CamonitorParserTest(unittest.TestCase):
  def createParser(self, aString, fieldSeparatorString=","):
    return CamonitorParser(aString, fieldSeparatorString)

  def parse(self, aString, fieldSeparatorString=","):
    return self.createParser(aString, fieldSeparatorString).parse()

  def testParse01(self):
    """Normal."""
    result = self.parse(
        "ioc23:val1,2012-09-27 19:55:51.952325,0,,\n"
        "ioc23:val2,<undefined>,0,UDF,INVALID")
    self.assertEqual(2, len(result))
    self.assertEqual("ioc23:val1", result[0].getPvName())
    self.assertEqual(0, result[0].getValue())
    self.assertEqual(datetime.datetime(2012, 9, 27, 19, 55, 51, 952325),
        result[0].getTimestamp())
    self.assertEqual(AlarmStatus.NO_ALARM, result[0].getStatus())
    self.assertEqual(AlarmSeverity.NO_ALARM, result[0].getSeverity())
    self.assertEqual("ioc23:val2", result[1].getPvName())
    self.assertEqual(0, result[1].getValue())
    self.assertEqual(None, result[1].getTimestamp())
    self.assertEqual(AlarmStatus.UDF, result[1].getStatus())
    self.assertEqual(AlarmSeverity.INVALID, result[1].getSeverity())

  def testParse02(self):
    """One PV found and one not found."""
    result = self.parse(
        "ioc23:val1,2012-09-27 19:55:51.952325,0,,\n"
        "ioc23:val3,*** Not connected (PV not found)")
    self.assertEqual("ioc23:val1", result[0].getPvName())
    self.assertEqual(0, result[0].getValue())
    self.assertEqual(datetime.datetime(2012, 9, 27, 19, 55, 51, 952325),
        result[0].getTimestamp())
    self.assertEqual(AlarmStatus.NO_ALARM, result[0].getStatus())
    self.assertEqual(AlarmSeverity.NO_ALARM, result[0].getSeverity())
    self.assertEqual("ioc23:val3", result[1].getPvName())
    self.assertEqual(None, result[1].getValue())
    self.assertEqual(None, result[1].getTimestamp())
    self.assertEqual(None, result[1].getStatus())
    self.assertEqual(None, result[1].getSeverity())

  def testParse03(self):
    """Empty program output."""
    result = self.parse("")
    self.assertEqual(0, len(result))

  def testParse04(self):
    """Two PVs not connected."""
    result = self.parse(
        "ioc23:val1,*** Not connected (PV not found)\n"
        "ioc23:val2,*** Not connected (PV not found)")
    self.assertEqual("ioc23:val1", result[0].getPvName())
    self.assertEqual(None, result[0].getValue())
    self.assertEqual(None, result[0].getTimestamp())
    self.assertEqual(None, result[0].getStatus())
    self.assertEqual(None, result[0].getSeverity())
    self.assertEqual("ioc23:val2", result[1].getPvName())
    self.assertEqual(None, result[1].getValue())
    self.assertEqual(None, result[1].getTimestamp())
    self.assertEqual(None, result[1].getStatus())
    self.assertEqual(None, result[1].getSeverity())

  def testParse05(self):
    """Lost PV connections and then reconnected."""
    result = self.parse(
        "ioc23:val1,<undefined>,0,UDF,INVALID\n"
        "ioc23:val2,<undefined>,0,UDF,INVALID\n"
        "CA.Client.Exception...............................................\n"
        "    Warning: \"Virtual circuit disconnect\"\n"
        "    Context: \"piano:5064\"\n"
        "    Source File: ../cac.cpp line 1214\n"
        "    Current Time: Thu Sep 27 2012 20:16:38.595780000\n"
        "..................................................................\n"
        "ioc23:val1,2012-09-27 20:16:38.595840 *** disconnected\n"
        "ioc23:val2,2012-09-27 20:16:38.595864 *** disconnected\n"
        "ioc23:val1,<undefined>,0,UDF,INVALID\n"
        "ioc23:val2,<undefined>,0,UDF,INVALID")
    self.assertEqual("ioc23:val1", result[0].getPvName())
    self.assertEqual(0, result[0].getValue())
    self.assertEqual(None, result[0].getTimestamp())
    self.assertEqual(AlarmStatus.UDF, result[0].getStatus())
    self.assertEqual(AlarmSeverity.INVALID, result[0].getSeverity())
    self.assertEqual("ioc23:val2", result[1].getPvName())
    self.assertEqual(0, result[1].getValue())
    self.assertEqual(None, result[1].getTimestamp())
    self.assertEqual(AlarmStatus.UDF, result[1].getStatus())
    self.assertEqual(AlarmSeverity.INVALID, result[1].getSeverity())
    self.assertEqual("ioc23:val1", result[2].getPvName())
    self.assertEqual(None, result[2].getValue())
    self.assertEqual(None, result[2].getTimestamp())
    self.assertEqual(None, result[2].getStatus())
    self.assertEqual(None, result[2].getSeverity())
    self.assertEqual("ioc23:val2", result[3].getPvName())
    self.assertEqual(None, result[3].getValue())
    self.assertEqual(None, result[3].getTimestamp())
    self.assertEqual(None, result[3].getStatus())
    self.assertEqual(None, result[3].getSeverity())
    self.assertEqual("ioc23:val1", result[4].getPvName())
    self.assertEqual(0, result[4].getValue())
    self.assertEqual(None, result[4].getTimestamp())
    self.assertEqual(AlarmStatus.UDF, result[4].getStatus())
    self.assertEqual(AlarmSeverity.INVALID, result[4].getSeverity())
    self.assertEqual("ioc23:val2", result[5].getPvName())
    self.assertEqual(0, result[5].getValue())
    self.assertEqual(None, result[5].getTimestamp())
    self.assertEqual(AlarmStatus.UDF, result[5].getStatus())
    self.assertEqual(AlarmSeverity.INVALID, result[5].getSeverity())

class CagetTest(unittest.TestCase):
  def testRun01(self):
    """Normal."""
    caput = Caput("ioc23:light1", "1")
    caput.setBlocking(True)
    caput.setTimeout(5.0)
    caput.run()
    caget = Caget(["ioc23:light1"])
    result = caget.run()
    self.assertEqual("ioc23:light1", result[0].getPvName())
    self.assertEqual(1, result[0].getValue())
    self.assertEqual(AlarmStatus.STATE, result[0].getStatus())
    self.assertEqual(AlarmSeverity.MINOR, result[0].getSeverity())
    caput = Caput("ioc23:light1", "0")
    caput.setBlocking(True)
    caput.setTimeout(5.0)
    caput.run()
    caget = Caget(["ioc23:light1"])
    result = caget.run()
    self.assertEqual("ioc23:light1", result[0].getPvName())
    self.assertEqual(0, result[0].getValue())
    self.assertEqual(AlarmStatus.NO_ALARM, result[0].getStatus())
    self.assertEqual(AlarmSeverity.NO_ALARM, result[0].getSeverity())

  def testRun02(self):
    """PV not found."""
    caget = Caget(["ioc23:nonexistent"])
    result = caget.run()
    self.assertEqual("ioc23:nonexistent", result[0].getPvName())
    self.assertEqual(None, result[0].getValue())
    self.assertEqual(None, result[0].getTimestamp())
    self.assertEqual(None, result[0].getStatus())
    self.assertEqual(None, result[0].getSeverity())

  def testRun03(self):
    """One PV found and one not found."""
    caget = Caget(["ioc23:light1", "ioc23:nonexistent"])
    result = caget.run()
    self.assertEqual("ioc23:light1", result[0].getPvName())
    self.assertEqual("ioc23:nonexistent", result[1].getPvName())

  def testRun04(self):
    """Normal, but with one PV name string rather than a list."""
    caput = Caput("ioc23:light1", "1")
    caput.setBlocking(True)
    caput.setTimeout(5.0)
    caput.run()
    caget = Caget("ioc23:light1")
    result = caget.run()
    self.assertEqual("ioc23:light1", result.getPvName())
    self.assertEqual(1, result.getValue())
    self.assertEqual(AlarmStatus.STATE, result.getStatus())
    self.assertEqual(AlarmSeverity.MINOR, result.getSeverity())
    caput = Caput("ioc23:light1", "0")
    caput.setBlocking(True)
    caput.setTimeout(5.0)
    caput.run()
    caget = Caget("ioc23:light1")
    result = caget.run()
    self.assertEqual("ioc23:light1", result.getPvName())
    self.assertEqual(0, result.getValue())
    self.assertEqual(AlarmStatus.NO_ALARM, result.getStatus())
    self.assertEqual(AlarmSeverity.NO_ALARM, result.getSeverity())

  def testRun05(self):
    """PV not found, but with one PV name string rather than a list."""
    caget = Caget("ioc23:nonexistent")
    result = caget.run()
    self.assertEqual("ioc23:nonexistent", result.getPvName())
    self.assertEqual(None, result.getValue())
    self.assertEqual(None, result.getTimestamp())
    self.assertEqual(None, result.getStatus())
    self.assertEqual(None, result.getSeverity())

class CaputTest(unittest.TestCase):
  def testRun01(self):
    """Normal."""
    caput = Caput("ioc23:light1", "1")
    result = caput.run()
    self.assertEqual("ioc23:light1", result.getPvName())
    self.assertEqual(1, result.getValue())
    self.assertEqual(AlarmStatus.STATE, result.getStatus())
    self.assertEqual(AlarmSeverity.MINOR, result.getSeverity())
    caput = Caput("ioc23:light1", "0")
    result = caput.run()
    self.assertEqual("ioc23:light1", result.getPvName())
    self.assertEqual(0, result.getValue())
    self.assertEqual(AlarmStatus.NO_ALARM, result.getStatus())
    self.assertEqual(AlarmSeverity.NO_ALARM, result.getSeverity())

  def testRun02(self):
    """Blocking put."""
    caput = Caput("ioc23:light1", "1")
    caput.setBlocking(True)
    caput.setTimeout(5.0)
    result = caput.run()
    self.assertEqual("ioc23:light1", result.getPvName())
    self.assertEqual(1, result.getValue())
    self.assertEqual(AlarmStatus.STATE, result.getStatus())
    self.assertEqual(AlarmSeverity.MINOR, result.getSeverity())
    caput = Caput("ioc23:light1", "0")
    caput.setBlocking(True)
    caput.setTimeout(5.0)
    result = caput.run()
    self.assertEqual("ioc23:light1", result.getPvName())
    self.assertEqual(0, result.getValue())
    self.assertEqual(AlarmStatus.NO_ALARM, result.getStatus())
    self.assertEqual(AlarmSeverity.NO_ALARM, result.getSeverity())

  def testRun03(self):
    """Write callback timeout due to default timeout (1.0)."""
    caput = Caput("ioc23:motorRqsPos1.ODLY", "2.0")
    caput.setBlocking(True)
    result = caput.run()
    try:
      caput = Caput("ioc23:motorRqsPos1.CALC", "23.0")
      caput.setBlocking(True)
      self.assertRaises(TimeoutError, caput.run)
    finally:
      # Set ODLY back to 0.0.  Must wait for this to complete with a timeout
      # that is longer than the previous ODLY value since the record
      # processing from the previous put has likely not finished yet (since we
      # timed out on it).  This also ensures that the record is no longer
      # processing when we leave this test case (which is important to avoid
      # adversely affecting other test cases that might use this record).
      caput = Caput("ioc23:motorRqsPos1.ODLY", "0.0")
      caput.setBlocking(True)
      caput.setTimeout(3.0)
      caput.run()

  def testRun04(self):
    """Write callback timeout due to timeout (2.0)."""
    caput = Caput("ioc23:motorRqsPos1.ODLY", "4.0")
    caput.setBlocking(True)
    result = caput.run()
    try:
      caput = Caput("ioc23:motorRqsPos1.CALC", "23.0")
      caput.setBlocking(True)
      caput.setTimeout(2.0)
      self.assertRaises(TimeoutError, caput.run)
    finally:
      # Set ODLY back to 0.0.  Must wait for this to complete with a timeout
      # that is longer than the previous ODLY value since the record
      # processing from the previous put has likely not finished yet (since we
      # timed out on it).  This also ensures that the record is no longer
      # processing when we leave this test case (which is important to avoid
      # adversely affecting other test cases that might use this record).
      caput = Caput("ioc23:motorRqsPos1.ODLY", "0.0")
      caput.setBlocking(True)
      caput.setTimeout(6.0)
      caput.run()

  def testRun05(self):
    """PV not found."""
    caput = Caput("ioc23:nonexistent", "1")
    result = caput.run()
    self.assertEqual("ioc23:nonexistent", result.getPvName())
    self.assertEqual(None, result.getValue())
    self.assertEqual(None, result.getTimestamp())
    self.assertEqual(None, result.getStatus())
    self.assertEqual(None, result.getSeverity())

class CamonitorTest(unittest.TestCase):
  def testRun01(self):
    """Normal."""
    camonitor = None
    try:
      caput = Caput("ioc23:light1", "0")
      caput.setBlocking(True)
      caput.setTimeout(5.0)
      caput.run()
      camonitor = Camonitor(["ioc23:light1"])
      camonitor.start()
      result = camonitor.next(5.0)
      self.assertEqual("ioc23:light1", result.getPvName())
      self.assertEqual(0, result.getValue())
      self.assertEqual(AlarmStatus.NO_ALARM, result.getStatus())
      self.assertEqual(AlarmSeverity.NO_ALARM, result.getSeverity())
      caput = Caput("ioc23:light1", "1")
      caput.run()
      result = camonitor.next(5.0)
      self.assertEqual("ioc23:light1", result.getPvName())
      self.assertEqual(1, result.getValue())
      self.assertEqual(AlarmStatus.STATE, result.getStatus())
      self.assertEqual(AlarmSeverity.MINOR, result.getSeverity())
    finally:
      if camonitor != None: camonitor.stop()

  def testRun02(self):
    """Only one PV, and it is not found."""
    camonitor = None
    try:
      camonitor = Camonitor(["ioc23:nonexistent"])
      camonitor.start()
      result = camonitor.next(5.0)
      self.assertEqual("ioc23:nonexistent", result.getPvName())
      self.assertEqual(None, result.getValue())
      self.assertEqual(None, result.getTimestamp())
      self.assertEqual(None, result.getStatus())
      self.assertEqual(None, result.getSeverity())
    finally:
      if camonitor != None: camonitor.stop()

  def testRunUntil01(self):
    """Normal."""
    caput = Caput("ioc23:light1", "0")
    caput.setBlocking(True)
    caput.setTimeout(5.0)
    caput.run()
    caput = Caput("ioc23:light2", "0")
    caput.setBlocking(True)
    caput.setTimeout(5.0)
    caput.run()
    caputThread = CaputSequenceThread([
        [None,           None, 1.0],
        ["ioc23:light1", 1,    1.0],
        ["ioc23:light2", 1,    1.0]])
    caputThread.start()
    camonitor = Camonitor(["ioc23:light1", "ioc23:light2"])
    result = camonitor.runUntil(
        [PvValueMatcher("ioc23:light1", 1),
            PvValueMatcher("ioc23:light2", 1)],
        10.0)
    caputThread.join()
    self.assertEqual("ioc23:light1", result[0].getPvName())
    self.assertEqual(1, result[0].getValue())
    self.assertEqual(AlarmStatus.STATE, result[0].getStatus())
    self.assertEqual(AlarmSeverity.MINOR, result[0].getSeverity())
    self.assertEqual("ioc23:light2", result[1].getPvName())
    self.assertEqual(1, result[1].getValue())
    self.assertEqual(AlarmStatus.STATE, result[1].getStatus())
    self.assertEqual(AlarmSeverity.MINOR, result[1].getSeverity())

  def testRunUntil02(self):
    """Only one PV, and it is not found."""
    camonitor = Camonitor(["ioc23:nonexistent"])
    result = camonitor.runUntil(
        [PvValueMatcher("ioc23:nonexistent")],
        10.0)
    self.assertEqual("ioc23:nonexistent", result[0].getPvName())
    self.assertEqual(None, result[0].getValue())
    self.assertEqual(None, result[0].getTimestamp())
    self.assertEqual(None, result[0].getStatus())
    self.assertEqual(None, result[0].getSeverity())

class PvTest(unittest.TestCase):
  def testRead01(self):
    """Normal."""
    caput = Caput("ioc23:light1", "0")
    caput.setBlocking(True)
    caput.setTimeout(5.0)
    caput.run()
    pv = Pv("ioc23:light1")
    pv.read()
    self.assertEqual("ioc23:light1", pv.getPvName())
    self.assertEqual(0, pv.getValue())
    self.assertEqual(AlarmStatus.NO_ALARM, pv.getStatus())
    self.assertEqual(AlarmSeverity.NO_ALARM, pv.getSeverity())
    caput = Caput("ioc23:light1", "1")
    caput.setBlocking(True)
    caput.setTimeout(5.0)
    caput.run()
    pv.read()
    self.assertEqual("ioc23:light1", pv.getPvName())
    self.assertEqual(1, pv.getValue())
    self.assertEqual(AlarmStatus.STATE, pv.getStatus())
    self.assertEqual(AlarmSeverity.MINOR, pv.getSeverity())

  def testWrite01(self):
    """Normal."""
    pv = Pv("ioc23:light1")
    pv.write(0)
    self.assertEqual(0, pv.getValue())
    self.assertEqual(AlarmStatus.NO_ALARM, pv.getStatus())
    self.assertEqual(AlarmSeverity.NO_ALARM, pv.getSeverity())
    pv.write(1)
    self.assertEqual(1, pv.getValue())
    self.assertEqual(AlarmStatus.STATE, pv.getStatus())
    self.assertEqual(AlarmSeverity.MINOR, pv.getSeverity())

  def testWriteAndWait01(self):
    """Normal."""
    odlyPv = Pv("ioc23:motorRqsPos1.ODLY")
    try:
      odlyPv.writeAndWait(1.0)
      pv = Pv("ioc23:motorRqsPos1.CALC")
      pv.writeAndWait(23.0)
      self.assertAlmostEqual(23.0, pv.getValue())
      self.assertEqual(AlarmStatus.NO_ALARM, pv.getStatus())
      self.assertEqual(AlarmSeverity.NO_ALARM, pv.getSeverity())
      pv.writeAndWait(2.5)
      self.assertEqual(2.5, pv.getValue())
      self.assertEqual(AlarmStatus.NO_ALARM, pv.getStatus())
      self.assertEqual(AlarmSeverity.NO_ALARM, pv.getSeverity())
    finally:
      odlyPv.writeAndWait(0.0)

if __name__ == "__main__": unittest.main()
